open GL
open FunGL
open Glut

type pos = {x:float; y:float; z:float}
type color = FunGL.rgb
type cube = {size:float; pos:pos; color:color}
type obj = Cube of cube
type app = {loc:pos; cnt:obj list}

let rand_color() =
  (Random.float 1.0,
   Random.float 1.0,
   Random.float 1.0)

let new_cube size ?(pos={x=0.; y=0.; z=0.}) color =
  Cube {size=size; pos=pos; color=color}

let init_app() =
  let init_pos = {x=0.; y=0.; z=0.} in
  {loc=init_pos; cnt=[]}

let draw_translated pos = draw_translated (pos.x, pos.y, pos.z) ;;

let display app =
  glClear [GL_COLOR_BUFFER_BIT];
  draw_as_identity (fun () ->
  draw_translated app.loc (fun () ->
    let display_item = function
    | Cube c ->
        draw_translated c.pos (fun () ->
        draw_with_rgb c.color (fun () ->
          glutWireCube c.size;
        ))
    in
    List.iter display_item app.cnt;
  ));
  glutSwapBuffers()

let app_add_x app step = {app with loc={app.loc with x= app.loc.x +. step}}
let app_add_y app step = {app with loc={app.loc with y= app.loc.y +. step}}
let app_add_z app step = {app with loc={app.loc with z= app.loc.z +. step}}

let app_sub_x app step = {app with loc={app.loc with x= app.loc.x -. step}}
let app_sub_y app step = {app with loc={app.loc with y= app.loc.y -. step}}
let app_sub_z app step = {app with loc={app.loc with z= app.loc.z -. step}}

let push_content app cnt = {app with cnt = cnt :: app.cnt}

let keyboard app ~key ~x ~y =
  match key with
  | 'q' | '\027' -> exit 0;
  | '-' -> glutPostRedisplay(); app_sub_z app 0.1
  | '+' -> glutPostRedisplay(); app_add_z app 0.1
  | _ -> app

let inv p = { x= -. p.x; y= -. p.y; z= -. p.z -. 3.0 }

let push_cube app =
  let cube = new_cube 0.1 ~pos:(inv app.loc) (rand_color()) in
  push_content app cube

let special app ~key ~x ~y =
  match key with
  | GLUT_KEY_F1 | GLUT_KEY_F2 | GLUT_KEY_F3 | GLUT_KEY_F4
  | GLUT_KEY_F5 | GLUT_KEY_F6 | GLUT_KEY_F7 | GLUT_KEY_F8
  | GLUT_KEY_F9 | GLUT_KEY_F10 | GLUT_KEY_F11 | GLUT_KEY_F12 -> app
  | GLUT_KEY_HOME | GLUT_KEY_PAGE_UP
  | GLUT_KEY_END  | GLUT_KEY_PAGE_DOWN -> app
  | GLUT_KEY_UP    -> glutPostRedisplay(); push_cube(app_sub_y app 0.1)
  | GLUT_KEY_DOWN  -> glutPostRedisplay(); push_cube(app_add_y app 0.1)
  | GLUT_KEY_LEFT  -> glutPostRedisplay(); push_cube(app_add_x app 0.1)
  | GLUT_KEY_RIGHT -> glutPostRedisplay(); push_cube(app_sub_x app 0.1)
  | GLUT_KEY_INSERT -> app

let () =
  Random.self_init();
  FunGlut.fun_glut
         ~display_mode:[GLUT_RGB; GLUT_DOUBLE]
         ~display ~keyboard ~special ~init:(init_app()) ()

